/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2011  <copyright holder> <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "ClearBall.h"
#include "simmaldribble.h"
#include <rcsc/geom/sector_2d.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_smart_kick.h>


using namespace rcsc;


bool ClearBall::execute( PlayerAgent * agent , Vector2D origin , double dist )
{
  const WorldModel & wm = agent->world();

  AngleDeg best_angle = 0.0;
  Vector2D kick_target = wm.self().pos();
  
  if ( wm.existKickableOpponent() )
  {
    //agent->doPointto( (wm.ball().pos() + Vector2D( 10.0, 0.0 )).x , (wm.ball().pos() + Vector2D( 10.0, 0.0 )).y );
    return Body_KickOneStep( wm.ball().pos() + Vector2D( 10.0, 0.0 ),
			     ServerParam::i().ballSpeedMax() ).execute( agent );
  }

  AngleDeg start = 0.0 , end = 0.0;
  bool clear_angle_bool = getClearAngle( agent , origin , dist , start , end );
  kick_target += Vector2D::polar2vector( 30.0, AngleDeg( ( start.degree() + end.degree() ) / 2 ) );

  if ( wm.gameMode().type() != GameMode::PlayOn )
  {
    //agent->doPointto( kick_target.x , kick_target.y );
    return Body_KickOneStep( kick_target , ServerParam::i().ballSpeedMax() ).execute( agent );
  }
  
  //agent->doPointto( kick_target.x , kick_target.y );
//   Vector2D first_vel = Vector2D::polar2vector( ServerParam::i().ballSpeedMax() , 
// 					     ( kick_target - ( wm.ball().pos() + wm.ball().pos() ) ).th() );
//   agent->doKick( ( first_vel - wm.ball().vel() ).r() / wm.self().kickRate() ,
// 		 ( kick_target - ( wm.ball().pos() + wm.ball().pos() ) ).th() );
//   return true;
  if( clear_angle_bool )
    return Body_SmartKick( kick_target,
			   ServerParam::i().ballSpeedMax(),
			   std::max( 2.5, ServerParam::i().ballSpeedMax() * 0.85 ),
			   2 ).execute( agent );
  return false;
}

/*!*********************************************************************************/
/*!
  clock wise
*/

bool ClearBall::getClearAngle( PlayerAgent* agent , Vector2D origin , double dist , 
			       AngleDeg & start , AngleDeg & end )
{
  const WorldModel & wm = agent->world();
  
  if( !origin.isValid() )
  {
    start = 0.0;
    end   = 0.0;
    return false;
  }
  
  std::vector<AngleDeg> angle_from_opps;
  
  double ang1 , ang2 =0;
  //Vector2D flag1 , flag2 = VectoD::INVALIDATED;
  Vector2D flag1 = Vector2D(-ServerParam::i().pitchHalfLength()+1,-ServerParam::i().pitchHalfWidth());
  Vector2D flag2 = Vector2D(-ServerParam::i().pitchHalfLength()+1,+ServerParam::i().pitchHalfWidth());
  ang1 = (flag1 - origin).th().degree();
  ang2 = (flag2 - origin).th().degree();
  ang1 = std::min(-90.0 , ang1);
  ang2 = std::max(+90.0 , ang2);
  //if(ang1!=0 && ang2!=0 )
    Sector2D sector( origin , 0.5 , dist , std::min(ang1,ang2) , std::max(ang1,ang2) );
  //else
   // Sector2D sector( origin , 0.5 , dist , -90.0 , +90 );
   
  
  const PlayerPtrCont & opps = wm.opponentsFromBall();
  
  PlayerPtrCont::const_iterator opps_end = opps.end();
  for ( PlayerPtrCont::const_iterator opp = opps.begin() ; opp != opps_end ; ++opp )
  {
    if( (*opp)->posCount() > 10 )
      continue;
	  
    if( (*opp)->isGhost() && (*opp)->posCount() >= 4 )
      continue;
	  
    if( sector.contains( (*opp)->pos() ) )
      angle_from_opps.push_back( (*opp)->rpos().th() );
  }
			   
  angle_from_opps.push_back( AngleDeg(ang1) );
  angle_from_opps.push_back( AngleDeg(ang2) );
  
  int angles_len = angle_from_opps.size();
  
  if( angles_len < 1 )
  {
    start = 0.0;
    end   = 0.0;
    return false;
  }
	  
  AngleDeg a = wm.self().body();
	  
  for( int j = angles_len - 1 ; j > 0 ; j-- )
    for( int k = 0 ; k < j ; k++ )  
      if( angle_from_opps.at(k).degree() > angle_from_opps.at(k+1).degree() )
      {
	a = angle_from_opps.at(k);
	angle_from_opps.at(k) = angle_from_opps.at(k+1);
	angle_from_opps.at(k+1) = a;
      }

//  std::sort( angle_from_opps.begin() , angle_from_opps.end() , AngleDeg::DegreeCmp() );
				
  std::vector<AngleDeg> dir_diff;
	  
  for( int i = 0 ; i < angles_len - 1 ; i++ )
   dir_diff.push_back( ( angle_from_opps.at(i) - angle_from_opps.at(i+1) ).abs() );

  if( dir_diff.size() < 1 )
  {
    start = 0.0;
    end   = 0.0;
    return false;
  }
  
  AngleDeg max_diff_dir_index = 0;
  int max_angle_index = -1;
	  
  for( int i = 0 ; i < angles_len -1 ; i++ )
    if ( angle_from_opps.at(i).degree() > max_diff_dir_index.degree() )
    {
      max_diff_dir_index = angle_from_opps.at(i);
      max_angle_index = i;
    }
  
  if( max_angle_index >= 0 )
  {
    start = angle_from_opps.at( max_angle_index );
    end   = angle_from_opps.at( max_angle_index + 1 );
//     return true;
  }

  const Vector2D my_pos = wm.self().pos();
  
  Vector2D ball_pos = wm.ball().pos();
  
  const Vector2D ball_vell = wm.ball().vel();
  
  int l = 1;
  
  AngleDeg dir = ( start.degree() + end.degree() ) / 2;
  
  Vector2D pos = my_pos + Vector2D::polar2vector( dist , dir );
  
  for( int x = 1 ; x <= ang2 , !SimMalDribble::isInLand( pos ) ; x += 2 )
  {
    int my_sign = sign( my_pos.y );
    
    pos = my_pos + Vector2D::polar2vector( dist , dir.degree() - my_sign * x );
  }
  
  for( int j = 0.0 , m = 0 ; m <= 90 ; j += ( 3 * l ) , m++ )
    for( l = 1 ; l >= -1 ; l -= 2 )
    {      
      const Vector2D required_ball_pos = pos - ball_pos;
      
      PlayerPtrCont::const_iterator opp = opps.begin();
      const PlayerPtrCont::const_iterator end = opps.end();
  
      const int time = SimMalDribble::cyclesObjectToReachPos( agent , *opp , pos , true ) 
		     - ( 1 - floor( required_ball_pos.r() / 14 ) );
		     
      const double term = ( 1.0 - std::pow( ServerParam::i().ballDecay(), time ) )
                        / ( 1.0 - ServerParam::i().ballDecay() );
      
      const Vector2D first_ball_pos = ball_pos + ball_vell;
      const Vector2D first_ball_vel = (required_ball_pos - wm.ball().rpos()) / term;
      
      for( int i = 0 ; i < time ; i++ )
      {
	double termprime = ( 1.0 - std::pow( ServerParam::i().ballDecay(), i ) ) 
			 / ( 1.0 - ServerParam::i().ballDecay() );
    
	ball_pos = first_ball_pos + (first_ball_vel * termprime);
    
	for ( opp = opps.begin() ; opp != end ; ++opp )
	{
    //       if( (*opp)->distFromSelf() <= 1.0 )
    // 	continue;
      
	  if( SimMalDribble::cyclesObjectToReachPos( agent , *opp , ball_pos , false ) <= i )
	  {
	    start = j;
	    return true;
	  }
	}
      }
    }
  
  start = 0.0;
  end   = 0.0;
  return false;
}

