// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_offensive_kick.h"
#include "bhv_pass.h"
#include "simmaldribble.h"
#include "strategy.h"
#include "taskevaluator.h"
#include "tasksequence.h"
#include "ClearBall.h"
#include "bhv_escape.h"

#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
Bhv_BasicOffensiveKick::execute( PlayerAgent * agent )
{   
   int role_num = Strategy::i().roleNumber( agent->world().self().unum() );
   
   std::vector<TaskSequence::TaskRoute> tasks;
   TaskEvaluator task_evaluator = TaskEvaluator( tasks );
   
   TaskSequence::TaskRoute fast_dribble;
   TaskSequence::TaskRoute dribble;
   TaskSequence::TaskRoute pass;
   TaskSequence::TaskRoute clear_ball;
   TaskSequence::TaskRoute hold_ball;
   TaskSequence::TaskRoute best_task;
   
   Vector2D P_target_point;
    double P_first_speed;
   int P_kick_step;
     
   bool can_pass = Bhv_Pass().virtualPassExecute( agent , P_target_point , P_first_speed );
   
   if( can_pass )
   {
 	pass.target_ = P_target_point;
 	pass.type_ = TaskSequence::PASS;
 	task_evaluator.pushEnd( pass );
   }
   
   Vector2D F_D_target_point;
   double F_D_dist_thr;
   double F_D_dash_power;
   int F_D_dash_count;
   bool F_D_dodge;
   bool F_D_speedy;
   bool F_D_Danger;
   bool can_fast_dribble = SimMalDribble().virtualFastDribbleExecute( agent ,
 								     F_D_target_point ,
 								     F_D_dist_thr , 
 								     F_D_dash_power , 
 								     F_D_dash_count , 
 								     F_D_dodge , 
 								     F_D_speedy , 
								     F_D_Danger );
   
   bool def = true;
   if( (role_num == 2 || role_num == 3) && agent->world().self().pos().x < -28.0 )
     def = false;
   
   if( can_fast_dribble && role_num != 2 && role_num != 3 )
   {
 	fast_dribble.target_ = F_D_target_point;
 	fast_dribble.type_ = TaskSequence::FAST_DRIBBLE;
 	task_evaluator.pushEnd( fast_dribble );
   }
   
   
   Vector2D D_target_point;
   double D_dist_thr;
   double D_dash_power;
   int D_dash_count;
   bool D_dodge;
   bool D_speedy; 
   bool D_danger;
   bool can_dribble = SimMalDribble().virtualDribbleExecute( agent , 
							     can_pass ,
 							    D_target_point ,
 							    D_dist_thr , 
 							    D_dash_power , 
 							    D_dash_count , 
 							    D_dodge , 
 							    D_speedy ,
							    D_danger );
   
   if( can_dribble && def )
   {
     dribble.target_ = D_target_point;
     dribble.type_ = TaskSequence::DRIBBLE;
     task_evaluator.pushEnd( dribble );
   }

   best_task = task_evaluator.execute( agent );
   
   if( best_task.type_ == TaskSequence::FAST_DRIBBLE )
   {
     SimMalDribble().dribbleExecute( agent , 
 				    F_D_target_point , 
 				    F_D_dist_thr , 
 				    F_D_dash_power , 
 				    F_D_dash_count , 
 				    F_D_dodge , 
 				    F_D_speedy ,
				    F_D_Danger );
     agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
     return true;
   }
   
   if( best_task.type_ == TaskSequence::DRIBBLE )
   {
     SimMalDribble().dribbleExecute( agent , 
 				    D_target_point , 
 				    D_dist_thr , 
 				    D_dash_power , 
 				    D_dash_count , 
 				    D_dodge , 
 				    D_speedy ,
				    D_danger );
 	agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
 	return true;
   }
   
   if( best_task.type_ == TaskSequence::PASS )
   {
     Bhv_Pass().passExecute( agent , P_target_point , P_first_speed );
     agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
     return true;
   }
   
   if( ( agent->world().self().pos().x > -20.0 ||
	 agent->world().opponentsFromSelf().front()->pos().dist( agent->world().self().pos() ) > 6.0 ) &&
 	 Body_HoldBall().execute( agent ) )
   {
     agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
     return true;
   }
   
   double dist = 30.0;
   if( agent->world().self().pos().x < -30.0 )
     dist = 20.0;
   
   if( ClearBall().execute( agent , agent->world().ball().pos() , dist ) )
   {
     agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
     return true;
   }
   
   if( Body_HoldBall().execute( agent ) )
   {
     agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
     return true;
   }
   
   return false;
}
