/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2012  Mohammadhossein Malmir <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "bhv_escape.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/math_util.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/composite_region_2d.h>

std::vector< Escape_Object > Bhv_Escape::cached_positions;
Escape_Object Bhv_Escape::M_cached_obj;
int Bhv_Escape::M_cached_obj_cycle;

#define DEBUG

bool Bhv_Escape::execute(rcsc::PlayerAgent* agent)
{
  const WorldModel & wm = agent->world();
  
  if( !wm.audioMemory().dribble().empty() )
  {
    if( wm.time().cycle() - wm.audioMemory().dribbleTime().cycle() <=
	wm.audioMemory().dribble().front().queue_count_  )
    {
      if( wm.audioMemory().dribble().front().sender_ < 6 )
	return false;
      
      if( wm.audioMemory().dribble().front().sender_ == 9 &&
	  wm.self().unum() == 10 )
	return false;
      
      if( wm.audioMemory().dribble().front().sender_ == 10 &&
	  wm.self().unum() == 9 )
	return false;
    }
  }
  
  if( wm.gameMode().type() != GameMode::PlayOn )
  {
    return false;
  }

  if( wm.self().stamina() < 1 )
  {
    return false;
  }
   
  if( wm.self().unum() < 9 || wm.self().unum() == 11 )
  {
    return false;
  }

  if( wm.ball().pos().x < -20.0 )
  {
    return false;
  }
  if( wm.existKickableOpponent() )
  {
    return false;
  }
  
  Vector2D best_pos = Vector2D::INVALIDATED;
  double best_score = -1000.0;

  if( get_best_pos(agent, best_pos, best_score) ) 
  {
    if( best_score > -1000.0 ) 
    {
      if( best_pos.dist( wm.self().pos() ) < 1.0 )
      {
	return true;
      }
      else
      {
	const Vector2D ball_pos = wm.ball().pos();
	double dash_power = Strategy::get_normal_dash_power( agent->world() );
	
	if( ball_pos.x >= 8.0 )
	  dash_power = ServerParam::i().maxDashPower();
	
        double point_thr = 0.0;
        
	if( ball_pos.x > 36.0 )
	  point_thr = 1.0;
        else
	  point_thr = 1.0;
	
        return Body_GoToPoint( best_pos, point_thr, dash_power ).execute( agent );
      }
    }
  }
    
  return false;  
}

///************************************************************************************/

bool Bhv_Escape::get_best_pos(PlayerAgent* agent, Vector2D& pos, double& score)
{
  double max_score = -10000.0;
  Escape_Object pos_tmp;
  const WorldModel & wm = agent->world();
       create_cache( agent );
  
  if( cached_positions.size() < 1 )
    return false;

  int n = 0;
  Vector2D best_pos = Vector2D::INVALIDATED;
  std::vector<Escape_Object>::iterator it_end = cached_positions.end();
  for ( std::vector<Escape_Object>::iterator it = cached_positions.begin() ; it != it_end ; it++, n++ ) 
      {
	if ((*it).Score() > max_score) 
	{
	  best_pos = (*it).Position();
	  max_score = (*it).Score();
	  pos_tmp = (*it);
	}
	 
      }
      
  pos = best_pos;
  score = max_score;
  M_cached_obj = pos_tmp;
  return true;
}

///************************************************************************************/

void Bhv_Escape::create_cache(PlayerAgent* agent)
{
  cached_positions.clear();
  const WorldModel & wm = agent->world();
  Escape_Object  pos_tmp;

  Vector2D home_pos = Strategy::i().getPosition( agent->world().self().unum() );
  Vector2D my_pos = agent->world().self().pos();
  
  std::vector<double> opps_y;
  const PlayerPtrCont & opps = wm.opponentsFromSelf();
  
  if( opps.size() < 1 )
    return;

  PlayerPtrCont::const_iterator last_opp = opps.end();
  for ( PlayerPtrCont::const_iterator opp = opps.begin() ; opp != last_opp ; ++opp )
  {
    if( (*opp)->goalie() )
      continue;
    
    if( (*opp)->pos().x < wm.offsideLineX() - 3.0 )
      continue;
    
    opps_y.push_back( (*opp)->pos().y );    
  }
  
  if( opps_y.size() < 1 )
    return;
  
  double a = 0.0;
  for( int i = opps_y.size() - 1 ; i > 0 ; i-- )
    for( int j = 0 ; j < i ; j++ )  
      if( opps_y.at(j) > opps_y.at(j+1) )
      {
	a = opps_y.at(j);
	opps_y.at(j) = opps_y.at(j+1);
	opps_y.at(j+1) = a;
      }
  
  int ball_y_sign = sign( wm.ball().pos().y );
  opps_y.push_back( ServerParam::i().pitchHalfWidth() * ball_y_sign );
  
  if( ball_y_sign > 0 )
    if( opps_y.front() > -ServerParam::i().goalHalfWidth() )
      opps_y.push_back( - ServerParam::i().goalHalfWidth() );
    
  if( ball_y_sign < 0 )
    if( opps_y.back() < ServerParam::i().goalHalfWidth() )
      opps_y.push_back( ServerParam::i().goalHalfWidth() );
    
  if( opps_y.size() < 1 )
    return;

  for( int i = opps_y.size() - 1 ; i > 0 ; i-- )
    for( int j = 0 ; j < i ; j++ )  
      if( opps_y.at(j) > opps_y.at(j+1) )
      {
	a = opps_y.at(j);
	opps_y.at(j) = opps_y.at(j+1);
	opps_y.at(j+1) = a;
      }
        
  std::vector<double> y_pos_vector;
  for( int i = 0 ; i < opps_y.size() - 1 ; i++ )
    y_pos_vector.push_back( ( opps_y.at(i) + opps_y.at(i+1) ) / 2 );
  
  std::vector<Vector2D> poses;
  for( int i = 0 ; i < y_pos_vector.size() ; i++ )
    poses.push_back( Vector2D( wm.offsideLineX() , y_pos_vector.at( i ) ) );
  
  int my_unum = wm.self().unum();
  
  Rect2D rect;
  
  switch (my_unum)
  {
    case 11 :
      rect = Rect2D( Vector2D( 0.0 ,-ServerParam::i().pitchHalfWidth() + 18.0 ) ,
		     Size2D( ServerParam::i().pitchHalfLength() , 
			     ServerParam::i().pitchHalfWidth() - 18.0 ) );
      break;
      
    case 10 :
      rect = Rect2D( Vector2D( 0.0 , ServerParam::i().pitchHalfWidth() / 3 ) ,
		     Size2D( ServerParam::i().pitchHalfLength() , 
	                     ServerParam::i().pitchHalfWidth() * 2 / 3 - 4.0 ) );
      break;
      
    case 9 :
      rect = Rect2D( Vector2D( 0.0 ,-ServerParam::i().pitchHalfWidth() ) ,
		     Size2D( ServerParam::i().pitchHalfLength() , 
			     ServerParam::i().pitchHalfWidth() * 2 / 3 - 4.0 ) );
      break; 
  }
  
  for( int i = 0 ; i < poses.size() ; i++ )
  {
    Vector2D pos = poses.at( i );
    double radius = std::max( 4.0 , fabs( opps_y.at(i) - opps_y.at(i+1) ) / 3 );
    Circle2D circle( pos , radius );
    
    for( int j = -radius ; j <= radius ; j += std::min( 1.0 , radius / 3 ) )
      for( int k = -radius ; k <= radius ; k += std::min( 1.0 , radius / 3 ) )
      {
	pos = pos + Vector2D( j , k );
	
	if( !rect.contains( pos ) )
	  continue; 
	
	if( !circle.contains( pos ) )
	  continue;
	
	pos_tmp = Escape_Object( agent, pos );
	cached_positions.push_back( pos_tmp );
      }
  }
}
