/*
 *    <one line to give the program's name and a brief idea of what it does.>
 *    Copyright (C) 2011  Mohammadhossein Malmir <mhmalmir@gmail.com>
 * 
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 * 
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef BHV_MARK_H
#define BHV_MARK_H

#include <rcsc/player/soccer_action.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/geom/vector_2d.h>
#include <vector>
#include <rcsc/common/server_param.h>

#include "strategy.h"
#include <rcsc/math_util.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/composite_region_2d.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_go_to_point.h>
#define MAX_THEIR_PLAYERS 11
#define MAXSIZE 25

using namespace rcsc;

class Bhv_Mark
: public SoccerBehavior
{
public:
  
  bool block_decision( PlayerAgent * agent ,
		       int & teammate_blocker );
  
  
  double area_danger( PlayerAgent * agent ,
		      PlayerObject * marker );
  
  static int last_blocker;
  static int last_marker_unum[7];
  static int last_say_cycle;

  class MarkEdge
  {
  private:
    
    const PlayerObject * marker;
    const PlayerObject * offender;
    PlayerAgent * agent;
    void sort_descending( std::vector<Vector2D> & vec_be_sorted , bool x_sort );
    
    int score;
    int offender_pos_count();
    int home_pos_dist();
    int offender_to_marker_dist();    
    int area_danger( PlayerAgent * agent );
    
    static const int        
    offender_pos_count_coeff 	     		= 1	,
    home_pos_dist_coeff	             		= -1	,
    offender_to_marker_dist_coeff 	     	= 1	,
    area_danger_coeff                    	= 1	;
    
    
  public:
    
    MarkEdge( const PlayerObject * marker_agent ,
	      const PlayerObject * offender_agent ,
	      PlayerAgent * self_agent ) 
    : marker(marker_agent) ,
      offender(offender_agent) ,
      agent(self_agent) 
    {
      score = offender_pos_count	() 		* offender_pos_count_coeff
      +       home_pos_dist		() 		* home_pos_dist_coeff
      +       offender_to_marker_dist	() 		* offender_to_marker_dist_coeff 	
      +       area_danger		( self_agent )	* area_danger_coeff;

    }
    
    int Score() 
    {
      return score; 
    }
    const PlayerObject * Marker() 
    {
      return marker; 
    }
    const PlayerObject * Offender() 
    {
      return offender; 
    }
    
    double U( double x );
    double gaussian( double x ,double mean ,double sigma );
   
  };

  double T( double x );
  double aussian( double x ,double mean , double sigma );
  static bool ball_our_possession;
  static int  cycle_ball_poss ;
  static const double score_thr = 0.0; //TODO
  int match[MAXSIZE];
  bool get_best_edge( PlayerAgent * agent );
  bool get_best_edge_simple( PlayerAgent * agent , int & offender , double * score );
  bool get_best_edge_mwbm( PlayerAgent * agent );  
  std::vector<MarkEdge> cached_edges;
  
  bool dfs( int v , int n , int adj[MAXSIZE][MAXSIZE] , int mark[] , int s[] , int t[] );
  bool matching( int n , int adj[MAXSIZE][MAXSIZE] , int s[] , int t[] , int sa[] );
  void getBipartiteGraph( int n , int m , int weight[MAXSIZE][MAXSIZE] );
  
  //static 
  
public:    

  int 		offenders_unum[MAX_THEIR_PLAYERS];
  bool  	execute( PlayerAgent * agent ){}
  bool 		markExecute( PlayerAgent * agent ,
			     int & opp_target_unum ,
			     bool & is_blocker );
  bool	        say ( PlayerAgent * agent ,
		      int blocker_unum );
  bool 		hear_n_parsing( PlayerAgent * agent ,
				int  & opp_target_unum ,
				bool & is_blocker );
};

#endif // BHV_MARK_H
